from typing import Any, Dict

from .config import Configuration
from .db.query_wrapper import QueryWrapper
from .db.update_wrapper import UpdateWrapper
from .db.insert_wrapper import InsertWrapper
from .db.wrapper import Wrapper
from .db.structures import structures
from seal.model.result import Results
from .cache import LRUCache
from loguru import logger

from .protocol.data_source_protocol import DataSourceProtocol


class Seal:

    def __init__(self):
        self._configuration = Configuration()
        self._initialized = False
        self.data_source_dict: Dict[str, DataSourceProtocol] = {}
        self._lru_cache: LRUCache = LRUCache(1024)

    def init(self, config_path):
        self._configuration.load(config_path)
        self._initialized = True

        logger.add(self.get_config('seal', 'loguru', 'path'),
                   rotation=self.get_config('seal', 'loguru', 'rotation'),
                   retention=self.get_config('seal', 'loguru', 'retention'),
                   level=self.get_config('seal', 'loguru', 'level'))

        data_source_config = self.get_config('seal', 'data_source')
        for data_source_name, data_source_conf in data_source_config.items():
            data_source = self.get_config('seal', 'data_source', data_source_name)
            if 'mysql' == data_source.get('type'):
                from .db.mysql.mysql_data_source import MysqlDataSource
                self.data_source_dict[data_source_name] = MysqlDataSource(name=data_source_name, conf=data_source_conf)
            elif 'sqlite' == data_source.get('type'):
                from .db.sqlite.sqlite_data_source import SqliteDataSource
                self.data_source_dict[data_source_name] = SqliteDataSource(name=data_source_name, conf=data_source_conf)
            else:
                raise ValueError(f'不支持的数据源类型: {data_source.get("type")}')

        if 'default' not in self.data_source_dict:
            raise ValueError('unknown default data source')
        logger.info(f'init seal with config: {config_path}')
        return self

    def data_source(self, data_source_name):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return self.data_source_dict.get(data_source_name)

    def get_config(self, *keys):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return self._configuration.get_conf(*keys)

    def get_config_default(self, *keys, default=None):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return self._configuration.get_conf_default(*keys, default=default)

    def query_wrapper(self, table: str, database: str | None = None, data_source: str = 'default') -> QueryWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()
        return QueryWrapper(table=table,
                            database=database,
                            data_source=self.data_source_dict[data_source],
                            tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                            tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                            logic_delete_field=self.get_config_default('seal', 'orm', 'logic_delete_field'),
                            logic_delete_true=self.get_config_default('seal', 'orm', 'logic_delete_true'),
                            logic_delete_false=self.get_config_default('seal', 'orm', 'logic_delete_false'),
                            )

    def update_wrapper(self, table: str, database: str | None = None, data_source: str = 'default') -> UpdateWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()
        return UpdateWrapper(table,
                             database=database,
                             data_source=self.data_source_dict[data_source],
                             tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                             tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                             update_by_field=self.get_config_default('seal', 'orm', 'update_by_field'),
                             update_at_field=self.get_config_default('seal', 'orm', 'update_at_field'),
                             logic_delete_field=self.get_config_default('seal', 'orm', 'logic_delete_field'),
                             logic_delete_true=self.get_config_default('seal', 'orm', 'logic_delete_true'),
                             logic_delete_false=self.get_config_default('seal', 'orm', 'logic_delete_false'),
                             )

    def insert_wrapper(self, table: str, database: str | None = None, data_source: str = 'default') -> InsertWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()
        return InsertWrapper(table,
                             database=database,
                             data_source=self.data_source_dict[data_source],
                             tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                             tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                             logic_delete_field=self.get_config_default('seal', 'orm', 'logic_delete_field'),
                             logic_delete_true=self.get_config_default('seal', 'orm', 'logic_delete_true'),
                             logic_delete_false=self.get_config_default('seal', 'orm', 'logic_delete_false'),
                             )

    # noinspection PyMethodMayBeStatic
    def conditions_wrapper(self) -> Wrapper:
        return Wrapper()

    def custom_query(self, data_source: str, sql: str, args=tuple[Any, ...]) -> Results:
        if data_source not in self.data_source_dict:
            raise ValueError(f'unknown data source: {data_source}')
        if sql is None:
            raise ValueError('sql is required')
        return self.data_source_dict[data_source].get_executor().custom_query(sql, args)

    def custom_update(self, data_source: str, sql: str, args=tuple[Any, ...]) -> int | None:
        if data_source not in self.data_source_dict:
            raise ValueError(f'unknown data source: {data_source}')
        if sql is None:
            raise ValueError('sql is required')
        return self.data_source_dict['default'].get_executor().custom_update(sql, args)

    # noinspection PyMethodMayBeStatic
    def get_structure(self, data_source='default', database=None, name: str = None) -> Any:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()
        if name is None:
            raise ValueError('name is required')
        return structures.get(data_source, database, name)

    def lru_cache(self) -> LRUCache:
        return self._lru_cache



import asyncio
import time
import re
from functools import wraps
import jwt
from fastapi import FastAPI, Request, Response, HTTPException, Depends
from starlette.middleware.cors import CORSMiddleware
from ..context import ctx_set
from ..model import Response as ResponseModel
from ..exception import BusinessException
from .. import seal
from loguru import logger


async def verify_token(request: Request = Request):
    if request.method == 'OPTIONS':
        return None
    path = request.url.path
    urls = seal.get_config('seal', 'authorization', 'excludes')
    for url in urls:
        # * -> [a-zA-Z0-9_\-]*, ** -> .*
        url = url.replace('*', '[a-zA-Z0-9_\-]*').replace('**', '.*')
        if re.match(url, path):
            return None
    try:
        payload = jwt.decode(request.headers['Authorization'],
                             seal.get_config('seal', 'authorization', 'jwt_key'),
                             algorithms=["HS256"])
        ctx_set({'uid': payload['uid']})
    except Exception as e:
        raise HTTPException(status_code=401, detail=f"Token is invalid: {e}")


app = FastAPI(dependencies=[Depends(verify_token)])


@app.middleware("http")
async def calc_time(request: Request, call_next):
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(process_time)
    return response


app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["*"]
)


@app.exception_handler(HTTPException)
async def http_exception_handler(request: Request, exc: HTTPException):
    return Response(status_code=exc.status_code, content=exc.detail)


def response_body(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            task = asyncio.create_task(func(*args, **kwargs))
            result = await task
            return ResponseModel.build(result).success()
        except BusinessException as e:
            logger.exception(e)
            return ResponseModel.build().error(message=e.message, code=e.code)
        except Exception as e:
            logger.exception(e)
            return ResponseModel.build().error(message=str(e))

    return wrapper


def get(path: str):
    def decorator(func):
        return app.get(path, response_model=ResponseModel)(response_body(func))

    return decorator


def post(path: str):
    def decorator(func):
        return app.post(path, response_model=ResponseModel)(response_body(func))

    return decorator


def delete(path: str):
    def decorator(func):
        return app.delete(path, response_model=ResponseModel)(response_body(func))

    return decorator


def put(path: str):
    def decorator(func):
        return app.put(path, response_model=ResponseModel)(response_body(func))

    return decorator



from contextvars import ContextVar

web_ctx = ContextVar('web', default={})


def ctx_get():
    return web_ctx.get()


def ctx_set(value: dict):
    web_ctx.set(value)


def ctx_uid():
    var = ctx_get()
    if 'uid' not in var:
        return None
    return var['uid']



from typing import Any, Tuple, List

from loguru import logger
from seal.model.result import Result, Results
from seal.protocol.data_source_protocol import DataSourceProtocol


class SqliteExecutor:

    def __init__(self, data_source: DataSourceProtocol):
        self.data_source = data_source

    def find(self, sql: str, args: Tuple[Any, ...], bean_type: Any) -> Result:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Result.empty()

            row = cursor.fetchone()
            if row is None:
                return Result.empty()

            return Result(row=row, bean_type=bean_type)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def find_list(self, sql: str, args: Tuple[Any, ...], bean_type: Any) -> Results:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows, bean_type=bean_type)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def count(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return None

            row = cursor.fetchone()
            if row is None:
                return None

            return row['COUNT(1)']
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def update(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def insert(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.lastrowid
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def insert_bulk(self, sql: str, args: List[Tuple[Any, ...]]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.executemany(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    # def insert_interator(self, data_iterator):
    #     connection = self.data_source.get_connection()
    #     cursor = connection.cursor()
    #     try:
    #         data_iterator(lambda sql, args: cursor.execute(sql, args))
    #         connection.commit()
    #         return cursor.rowcount
    #     except Exception as e:
    #         logger.exception(e)
    #         raise e
    #     finally:
    #         cursor.close()
    #         connection.close()

    def custom_query(self, sql: str, args: Tuple[Any, ...]) -> Results:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def custom_update(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()



from typing import Any, List, Tuple

from .structures import structures
from .wrapper import Wrapper
from .sql_builder import build_select, build_count
from seal.model.result import Result, Results
from dataclasses import fields

from ..protocol.data_source_protocol import DataSourceProtocol


class QueryWrapper(Wrapper):
    def __init__(self,
                 table: str,
                 database=None,
                 data_source: DataSourceProtocol = None,
                 tenant_field=None,
                 tenant_value=None,
                 logic_delete_field=None,
                 logic_delete_true=None,
                 logic_delete_false=None):
        super().__init__(tenant_field=tenant_field,
                         tenant_value=tenant_value,
                         logic_delete_field=logic_delete_field,
                         logic_delete_true=logic_delete_true,
                         logic_delete_false=logic_delete_false)
        self.table = table
        if database is not None:
            self.table = f'{database}.{table}'
        self.data_source = data_source

        self.result_type = structures.get(data_source=self.data_source.get_name(), database=database, table=table)
        if self.result_type is None:
            self.result_type = self.data_source.load_structure(database, table)
            structures.register(data_source=self.data_source.get_name(), database=database, table=table, structure=self.result_type)

        self.limit_ = None
        self.offset = None
        self.order_by = None
        self.field_list = []
        self.ignore_fields = []

    def select(self, *field_list) -> 'QueryWrapper':
        self.field_list = field_list
        return self

    def ignore(self, *field_list) -> 'QueryWrapper':
        self.ignore_fields = field_list
        return self

    def sort(self, *order_by) -> 'QueryWrapper':
        self.order_by = order_by
        return self

    def limit(self, limit: int) -> 'QueryWrapper':
        self.limit_ = limit
        return self

    def offset(self, offset: int) -> 'QueryWrapper':
        self.offset = offset
        return self

    def one(self, as_dict=False, **options) -> Any:
        sql, args = self.build_statement(**options)
        result: Result = self.data_source.get_executor().find(sql, args, self.result_type)
        if as_dict:
            return result.as_dict()
        return result.get()

    def one_dict(self, **options) -> dict:
        return self.one(as_dict=True, **options)

    def list(self, as_dict=False, **options) -> List[Any]:
        sql, args = self.build_statement(**options)
        results: Results = self.data_source.get_executor().find_list(sql, args, self.result_type)
        if as_dict:
            return results.as_dict()
        return results.get()

    def list_dict(self, **options) -> List[dict]:
        return self.list(as_dict=True, **options)

    def page(self, page: int, page_size: int, as_dict=False, **options) -> Tuple[List[Any], int]:
        self.limit_ = page_size
        self.offset = (page - 1) * page_size
        sql, args = self.build_statement(**options)
        results: Results = self.data_source.get_executor().find_list(sql, args, self.result_type)
        count = self.count()
        if as_dict:
            return results.as_dict(), count
        return results.get(), count

    def page_dict(self, page: int, page_size: int, **options) -> Tuple[List[dict], int]:
        return self.page(page, page_size, as_dict=True, **options)

    def count(self):
        sql, args = build_count(self)
        return self.data_source.get_executor().count(sql, args)

    def build_statement(self, **options) -> Tuple[str, Tuple[Any, ...]]:
        self.handle_public_fields(**options)

        if len(self.field_list) == 0:
            self.field_list = [field.name for field in fields(self.result_type) if field.name not in self.ignore_fields]
        return build_select(self)

    def build_sql(self, **options) -> str:
        sql, args = self.build_statement(**options)
        for arg in args:
            if isinstance(arg, str):
                sql = sql.replace('?', f"'{arg}'", 1)
            else:
                sql = sql.replace('?', str(arg), 1)
        return sql



from typing import Protocol, Any

from .executor_protocol import ExecutorProtocol


class DataSourceProtocol(Protocol):
    def get_name(self) -> str:
        ...

    def get_type(self) -> str:
        ...

    def get_executor(self) -> ExecutorProtocol:
        ...

    def load_structure(self, database: str, table: str) -> Any:
        ...

    def get_connection(self):
        ...

    def get_default_database(self) -> str:
        ...
